#ifndef _MOVIE_H
#define _MOVIE_H

#include "movie-driver.h"
#include "state.h"
void MDFNMOV_AddJoy(SFORMAT *sf);
void MDFNMOV_CheckMovies(void);
void MDFNMOV_Stop(void);
void MDFNMOV_AddCommand(int cmd);
bool MDFN_DrawMovies(uint32 *, const MDFN_Rect *OSDRect, uint32 pitch32);
bool MDFNMOV_IsPlaying(void);
bool MDFNMOV_IsRecording(void);
void MDFNMOV_ForceRecord(StateMem *sm);
StateMem MDFNMOV_GrabRewindJoy(void);
void MDFNMOV_RecordState();
#endif
